-- postgres 9.3

------------------------------------------------------------------------
-- PostgreSQL CMS
------------------------------------------------------------------------

START TRANSACTION;

------------------------------------------------------------------------
-- init db
------------------------------------------------------------------------


CREATE EXTENSION IF NOT EXISTS pgcrypto WITH SCHEMA pg_catalog;

DROP SCHEMA IF EXISTS interface CASCADE;
CREATE SCHEMA interface;
COMMENT ON SCHEMA interface IS 'functions which are exported over HTTP';

DROP SCHEMA IF EXISTS private CASCADE;
CREATE SCHEMA private;
COMMENT ON SCHEMA private IS 'private data, functions';

DROP SCHEMA IF EXISTS session CASCADE;
CREATE SCHEMA session;
COMMENT ON SCHEMA session IS '(temporary) session data';

DROP SCHEMA IF EXISTS log CASCADE;
CREATE SCHEMA log;
COMMENT ON SCHEMA log IS 'log data';

DROP SCHEMA IF EXISTS content CASCADE;
CREATE SCHEMA content;
COMMENT ON SCHEMA content IS 'all our content.';

DROP SCHEMA IF EXISTS security CASCADE;
CREATE SCHEMA security;
COMMENT ON SCHEMA security IS 'security';


------------------------------------------------------------------------
-- accounts & sessions
------------------------------------------------------------------------


CREATE TABLE private.account (
  id
	UUID
	NOT NULL
	DEFAULT encode(gen_random_bytes(16), 'hex')::UUID
	PRIMARY KEY,
  username
	VARCHAR(30)
	NOT NULL
	CHECK(username ~ '^[_a-z]{1}[-_a-z0-9]{0,29}$'),
  salt
	TEXT
	NOT NULL
	DEFAULT gen_salt('bf'),
  passhash
	TEXT
	NOT NULL
	DEFAULT ''
);

CREATE UNIQUE INDEX account_idx ON private.account(username);

CREATE TABLE session.session (
  id
    UUID
    NOT NULL
    DEFAULT encode(gen_random_bytes(16), 'hex')::UUID
    PRIMARY KEY,
  account
    UUID
    NOT NULL
    REFERENCES private.account(id),
  started
    TIMESTAMP WITH TIME ZONE
    NOT NULL
    DEFAULT CURRENT_TIMESTAMP,
  last_touch
    TIMESTAMP WITH TIME ZONE
    NOT NULL
    DEFAULT CURRENT_TIMESTAMP
);

-- returns the account of a session
CREATE OR REPLACE FUNCTION private.session_get_account(
	IN session UUID,
	OUT account UUID) AS $$
BEGIN
	SELECT session.account
	FROM session.session
	WHERE id = $1::UUID
	INTO account;
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION private.is_self(
	IN username TEXT,
	IN session UUID,
	OUT ok BOOL) AS $$
BEGIN
	ok := false;
	SELECT account.id IS NOT NULL
	FROM session.session, private.account
	WHERE
		session.id = $2 AND
		session.account = account.id AND
		account.username = $1
	INTO STRICT ok;

	EXCEPTION
		WHEN no_data_found THEN
		NULL;
END;
$$ LANGUAGE plpgsql;


------------------------------------------------------------------------
-- administrators
------------------------------------------------------------------------


-- entries denote administrator rights are granted.
CREATE TABLE private.administrator (
  account
	UUID
	REFERENCES private.account(id) ON DELETE CASCADE
	PRIMARY KEY
);

-- return true if session has an admin account
CREATE OR REPLACE FUNCTION private.is_admin(
	IN session UUID,
	OUT ok BOOL) AS $$
BEGIN
	ok := FALSE;
	SELECT administrator.account IS NOT NULL
	FROM session.session, private.administrator
	WHERE
		session.id = $1 AND
		session.account = administrator.account
	INTO STRICT ok;

	EXCEPTION
	WHEN no_data_found THEN
		NULL;
END;
$$ LANGUAGE plpgsql;

-- add a username as admin
CREATE OR REPLACE FUNCTION private.grant_admin(
	IN username TEXT,
	OUT ok BOOL,
	OUT reason TEXT) AS
$$
BEGIN
	ok := FALSE;
	BEGIN
		WITH ins AS (
				INSERT INTO private.administrator (account)
				SELECT id FROM private.account 
				WHERE account.username = $1
				RETURNING 1
		)
		SELECT * FROM ins INTO STRICT ok;
		reason = 'admin added';
	EXCEPTION
		WHEN unique_violation THEN
			reason := 'already admin';
			ok := TRUE;
			RETURN;
		WHEN no_data_found THEN
			reason := 'no such user';
			ok := FALSE;
			RETURN;
	END;
END;
$$ LANGUAGE plpgsql;

-- revoke someone's admin rights
CREATE OR REPLACE FUNCTION private.revoke_admin(
	IN username TEXT,
	OUT ok BOOL,
	OUT reason TEXT) AS
$$
BEGIN
	ok := FALSE;
	DELETE FROM private.administrator
	USING private.account
	WHERE administrator.account = account.id
		AND account.username = $1;
	IF NOT FOUND THEN
		reason = 'not an administrator';
		RETURN;
	END IF;
	ok = TRUE;
END;
$$ LANGUAGE plpgsql;


------------------------------------------------------------------------
-- logging
------------------------------------------------------------------------


CREATE TABLE log.event_type (
	id
		SERIAL
		NOT NULL
		PRIMARY KEY,
	name 
		TEXT
		NOT NULL,
	message
		TEXT
		NOT NULL,
	numargs
		INT
		NOT NULL
		DEFAULT 0
);

CREATE TABLE log.event (
	id
		SERIAL
		NOT NULL
		PRIMARY KEY,
	account
		UUID
		NULL
		REFERENCES private.account(id),
	event_type
		INT
		NOT NULL
		REFERENCES log.event_type(id),
	args
		TEXT[]
		NOT NULL,
	time
		TIMESTAMP WITH TIME ZONE
		NOT NULL
		DEFAULT CURRENT_TIMESTAMP
);

-- log an event
CREATE OR REPLACE FUNCTION private.log(
	IN name TEXT,
	VARIADIC args TEXT[]) RETURNS VOID AS
$$
DECLARE
	event_type INT;
BEGIN
	SELECT event_type.id FROM log.event_type 
	INTO event_type 
	WHERE log.event_type.name = $1;

	INSERT INTO log.event (event_type, args)
	VALUES (event_type, args);
END;
$$ LANGUAGE plpgsql;

-- format an event as string
CREATE OR REPLACE FUNCTION private.log_string(
	IN id INT,
	OUT result TEXT) AS
$$
BEGIN
	SELECT format(event_type.message, VARIADIC event.args) 
	INTO result
	FROM log.event, log.event_type
	WHERE event.id = $1;
END;
$$ LANGUAGE plpgsql;


------------------------------------------------------------------------
-- content
------------------------------------------------------------------------


-- actual content (eg a photo or an album) will inherit from this.
CREATE TABLE content.base (
	id
		UUID
		NOT NULL
		DEFAULT encode(gen_random_bytes(16), 'hex')::UUID
		PRIMARY KEY,
	author
		UUID
		NOT NULL
		REFERENCES private.account(id) ON DELETE CASCADE,
	created
		TIMESTAMP WITH TIME ZONE
		NOT NULL
		DEFAULT CURRENT_TIMESTAMP,
	type
		VARCHAR(8)
		NOT NULL
);


------------------------------------------------------------------------
-- security
------------------------------------------------------------------------


CREATE TABLE security.action_types (
	id
		SERIAL
		NOT NULL
		PRIMARY KEY,
	name
		-- example: stream:query, stream:delete, poll:vote, ...
		TEXT
		NOT NULL
		UNIQUE
);

CREATE TABLE security.grants (
	id
		UUID
		NOT NULL
		DEFAULT encode(gen_random_bytes(16), 'hex')::UUID
		PRIMARY KEY,
	parent_id
		UUID
		-- NULL if grant created by author of content
		NULL
		REFERENCES security.grants(id) ON DELETE CASCADE,
	from_account
		UUID
		NOT NULL
		REFERENCES private.account(id) ON DELETE CASCADE,
	to_account
		UUID
		NOT NULL
		REFERENCES private.account(id) ON DELETE CASCADE,
	content
		UUID
		NOT NULL
		REFERENCES content.base(id) ON DELETE CASCADE,
	action_type
		INT
		NOT NULL
		REFERENCES security.action_types(id)
);

-- user_has_permission('stream:post', account, stream)
-- if stream is owned by account
-- 	return true
-- else 
-- 	SELECT * FROM stream_grants JOIN grants JOIN grant_type 
-- 	WHERE action = $1 && stream = $2 && grants.to_account = $3 

-- user_grant_permission('stream:post', from, to, stream)
-- if from owns stream,
-- 	insert with empty paren_id
-- else error
-- a user might be granted access to a stream by many users, 
-- we need to specify which grant we wish to pass on.

-- Does username correspond to session id?
------------------------------------------------------------------------


------------------------------------------------------------------------


-- Does username correspond to session id?
CREATE OR REPLACE FUNCTION
private.has_permission(
	IN username TEXT,
	IN session UUID,
	OUT ok BOOL) AS
$$
BEGIN
	ok := false;
	SELECT account.id IS NOT NULL
	FROM session.session, private.account
	WHERE
		session.id = $2 AND
		session.account = account.id AND
		account.username = $1
	INTO STRICT ok;

	EXCEPTION
		WHEN no_data_found THEN
		NULL;
END;
$$ LANGUAGE plpgsql;


------------------------------------------------------------------------
-- public interface
------------------------------------------------------------------------


-- register a new account
CREATE OR REPLACE FUNCTION interface.register(
	IN username TEXT,
	IN password TEXT,
	OUT ok BOOL,
	OUT reason TEXT ) AS
$$
DECLARE
	salt TEXT;
BEGIN
	salt := gen_salt('bf');

	INSERT INTO private.account (username, salt, passhash)
	VALUES ($1, salt, crypt(password, salt));
	ok := true;

	EXCEPTION
	WHEN unique_violation THEN
		reason := format('register: username %s taken', username);
		ok := false;
	WHEN check_violation THEN
		reason := format('register: username %s has wrong format', username);
		ok := false;
END;
$$
LANGUAGE plpgsql;

-- create a session logged in to an account
CREATE OR REPLACE FUNCTION interface.login (
	IN username TEXT,
	IN password TEXT,
	OUT ok BOOL,
	OUT reason TEXT,
	OUT session UUID,
	OUT admin BOOL) AS
$$
DECLARE
	v_account UUID;
	v_session UUID;
BEGIN
	ok := false;

	SELECT account.id, account IS NOT NULL
	INTO v_account, admin
	FROM private.account LEFT OUTER JOIN private.administrator
		ON ( account.id = administrator.account )
	WHERE
		account.username = $1 AND
		account.passhash = crypt(password, account.salt);

	IF NOT FOUND THEN
		reason := format('login: username % not found', username);
		RETURN;
	END IF;

	v_session := ENCODE( gen_random_bytes( 16 ), 'hex' )::UUID;
	INSERT INTO session.session (id, account) VALUES (v_session, v_account);

	ok := true;
	session = v_session;
END;
$$
LANGUAGE plpgsql;

-- destroys a session logged into an account.
CREATE OR REPLACE FUNCTION interface.logout(
	session VARCHAR(36),
	OUT ok BOOL,
	OUT reason TEXT) AS
$$
DECLARE
	v_session UUID;
BEGIN
	v_session := session::UUID;
	ok := true;

	DELETE FROM session.session
	WHERE session.id = v_session;

	IF NOT FOUND THEN
		reason := format('logout: session %s was not found', v_session);
		ok := false;
	END IF;
END;
$$
LANGUAGE plpgsql;

-- change the password on an account.
CREATE OR REPLACE FUNCTION interface.set_password(
	IN username TEXT,
	IN password TEXT,
	IN session TEXT,
	OUT ok BOOL,
	OUT reason TEXT) AS $$
DECLARE
	v_session UUID;
BEGIN
	v_session = session::UUID;
	ok := false;
	IF private.is_self(username, v_session) OR private.is_admin(v_session) THEN
		UPDATE private.account
		SET passhash = crypt(password, account.salt)
		WHERE account.username = $1;
		ok = true;
	ELSE
		reason := 'unauthorized user';
	END IF;
END;
$$ LANGUAGE plpgsql;

-- get a list of usernames in the system
CREATE OR REPLACE FUNCTION interface.get_usernames(
	IN session TEXT,
	OUT usernames VARCHAR(30)[],
	OUT ok BOOL,
	OUT reason TEXT) AS $$
BEGIN
	ok := false;
	IF private.is_admin(session::UUID) THEN
		SELECT array_agg(username)
		INTO usernames
		FROM private.account;
		ok := true;
	ELSE
		reason := 'get_usernames: access denied';
	END IF;
END;
$$ LANGUAGE plpgsql;


------------------------------------------------------------------------
-- some tests
------------------------------------------------------------------------


SELECT 'register username', * FROM interface.register('username', 'password');
SELECT 'make username admin', * FROM private.grant_admin('username');
SELECT 'register foo', * FROM interface.register('foo', 'bar');
SELECT 'register jim', * FROM interface.register('jim', 'bar');
SELECT 'register jane', * FROM interface.register('jane', 'bar');
SELECT 'make jim admin', * FROM private.grant_admin('jim');
SELECT 'make jim admin (again)', * FROM private.grant_admin('jim');
SELECT 'make jimz admin', * FROM private.grant_admin('jimz');
SELECT 'revoke jim admin', * FROM private.revoke_admin('jim');
SELECT 'revoke jim admin (again)', * FROM private.revoke_admin('jim');


------------------------------------------------------------------------

COMMIT;
